"""List events for a finetuning run"""

from typing import Union

from mcli import list_finetuning_events

from databricks_genai.api.config import configure_request
from databricks_genai.types.common import ObjectList
from databricks_genai.types.finetuning import FinetuningEvent, FinetuningRun


@configure_request
def get_events(finetuning_run: Union[str, FinetuningRun]) -> ObjectList[FinetuningEvent]:
    """List finetuning runs
    
    Args:
        finetuning_run (Union[str, FinetuningRun]): The finetuning run to get events for.

    Returns:
        List[FinetuningEvent]: A list of finetuning events. Each event has an event 
            type, time, and message.
    """
    finetune = finetuning_run if isinstance(finetuning_run, str) else finetuning_run.name
    events = list_finetuning_events(finetune)
    return ObjectList(events, FinetuningEvent)



"""List multiple finetuning runs"""

from datetime import datetime
from typing import List, Optional, Union

from mcli import get_finetuning_runs

from databricks_genai.api.config import configure_request
from databricks_genai.types.common import ObjectList
from databricks_genai.types.finetuning import FinetuningRun


@configure_request
def list(  # pylint: disable=redefined-builtin
    finetuning_runs: Optional[Union[List[str], List[FinetuningRun]]] = None,
    *,
    user_emails: Optional[List[str]] = None,
    before: Optional[Union[str, datetime]] = None,
    after: Optional[Union[str, datetime]] = None,
    limit: Optional[int] = 50,
) -> ObjectList[FinetuningRun]:
    """List one or many finetuning runs
    
    Args:
        finetuning_runs (Optional[Union[List[str], List[FinetuningRun]]], optional): A list of
            finetuning runs to get. Defaults to selecting all finetuning runs.
        user_emails (Optional[List[str]], optional): A list of user emails to filter by.
            Defaults to no user filter.
        before (Optional[Union[str, datetime]], optional): A datetime or datetime string to
            filter finetuning runs before. Defaults to all finetuning runs.
        after (Optional[Union[str, datetime]], optional): A datetime or datetime string to
            filter finetuning runs after. Defaults to all finetuning runs.
    
    Returns:
        ObjectList[FinetuningRun]: A list of finetuning runs, max 50
    """

    return ObjectList(
        get_finetuning_runs(
            finetuning_runs=finetuning_runs,
            user_emails=user_emails,
            before=before,
            after=after,
            include_details=True,
            limit=limit,
        ), FinetuningRun)



"""Utils for finetuning API"""
import json
import os
from typing import Optional, Union

from databricks.sdk import WorkspaceClient
from datasets import get_dataset_split_names
from mcli.api.exceptions import ValidationError
from mlflow import MlflowClient

_UC_VOLUME_LIST_API_ENDPOINT = '/api/2.0/fs/list'


def validate_model_registry_path(path: str) -> None:
    split_path = path.split('.')
    if len(split_path) == 2:
        catalog, schema_name = split_path
    elif len(split_path) == 3:
        catalog, schema_name, _ = split_path
    else:
        raise ValidationError(f'model_registry_path must be in the format '
                              f'catalog.schema or catalog.schema.model_name, but got {path}')
    for component in split_path:
        if len(component) == 0:
            raise ValidationError(f'model_registry_path must be in the format '
                                  f'catalog.schema or catalog.schema.model_name, but got {path}')
    validate_catalog_schema(catalog, schema_name, "model_registry_path")


def validate_delta_table(path: str, input_type: str) -> None:
    split_path = path.split('.')
    if len(split_path) != 3:
        raise ValidationError(f'Delta table input to {input_type} must be in the format '
                              f'catalog.schema.table, but got {path}.')
    for component in split_path:
        if len(component) == 0:
            raise ValidationError(f'Delta table input to {input_type} must be in the format '
                                  f'catalog.schema.table, but got {path}.')
    catalog, schema, _ = split_path
    validate_catalog_schema(catalog, schema, input_type)


def validate_catalog_schema(catalog: str, schema_name: str, input_type: str) -> None:
    w = WorkspaceClient()
    schemas = w.schemas.list(catalog)
    if schema_name not in [schema.name for schema in schemas]:
        raise ValidationError(f'Failed to find schema "{schema_name}" in catalog "{catalog}". Please make sure '
                              f'your {input_type} is valid and exists in the Unity Catalog.')


def validate_experiment_path(experiment_path: str) -> None:
    try:
        client = MlflowClient(tracking_uri='databricks')
        experiment = client.get_experiment_by_name(experiment_path)
        if not experiment:
            client.create_experiment(experiment_path)
    except Exception as e:
        raise ValidationError(f'Failed to get or create MLflow experiment {experiment_path}. Please make sure '
                              'your experiment name is valid.') from e


def validate_uc_path(uc_path: str) -> None:
    if not uc_path.startswith("dbfs:/Volumes"):
        raise ValidationError('Databricks Unity Catalog Volumes paths should start with "dbfs:/Volumes".')
    path = os.path.normpath(uc_path[len("dbfs:/"):])
    dirs = path.split(os.sep)
    if len(dirs) < 4:
        raise ValidationError(f'Databricks Unity Catalog Volumes path expected to start with ' \
            f'`dbfs:/Volumes/<catalog-name>/<schema-name>/<volume-name>`. Found path={uc_path}')
    object_path = "/" + path
    client = WorkspaceClient()
    if path.endswith(".jsonl"):
        try:
            client.files.get_status(object_path)
        except Exception as e:
            raise ValidationError(f"Failed to access Unity Catalog path {uc_path}.") from e
    else:
        data = json.dumps({'path': object_path})
        try:
            resp = client.api_client.do(method='GET',
                                        path=_UC_VOLUME_LIST_API_ENDPOINT,
                                        data=data,
                                        headers={'Source': 'mosaicml/finetuning'})
        except Exception as e:
            raise ValidationError(f"Failed to access Unity Catalog path {uc_path}.") from e
        if len([f['path'] for f in resp.get('files', []) if not f['is_dir']]) == 0:
            raise ValidationError(
                f"No files found in Unity Catalog path {uc_path}. Please make sure your path includes input data.")


def validate_hf_dataset(dataset_name_with_split: str) -> None:
    print(f"Assuming {dataset_name_with_split} is a Hugging Face dataset (not in format `dbfs:/Volumes` or "
          "`/Volumes`). Validating...")
    split_dataset_name = dataset_name_with_split.split('/')
    if len(split_dataset_name) < 2:
        raise ValidationError(
            f"Hugging Face dataset {dataset_name_with_split} must be in the format <dataset>/<split> or "
            "<entity>/<dataset>/<split>.")
    dataset_name, split = '/'.join(split_dataset_name[0:-1]), split_dataset_name[-1]
    try:
        splits = get_dataset_split_names(dataset_name)
    except Exception as e:
        raise ValidationError(
            f"Failed to access Hugging Face dataset {dataset_name_with_split}. Please make sure that the split "
            "is valid and that your dataset does not have subsets.") from e
    if split not in splits:
        raise ValidationError(f"Failed to access Hugging Face dataset {dataset_name_with_split}. Split not found.")
    print("Hugging Face dataset validation successful.")


def validate_data_prep(train_data_path: str, data_prep_config: dict, eval_data_path: Optional[str] = None):
    # TODO: actually use these arguments
    # Assumed HF dataset
    if len(train_data_path.split('/')) > 0:
        return
    if len(train_data_path.split('.')) >= 3 and len(data_prep_config.keys()) == 0:
        raise ValidationError("Train data path was a Delta table, but data prep config was not provided.")
    if len(eval_data_path.split('/')) > 0:
        return
    if len(eval_data_path.split('.')) >= 3 and len(data_prep_config.keys()) == 0:
        raise ValidationError("Eval data path was a Delta table, but data prep config was not provided.")


def validate_create_finetuning_run_inputs(train_data_path: str,
                                          model_registry_path: Optional[str] = None,
                                          experiment_path: Optional[str] = None,
                                          eval_data_path: Optional[str] = None,
                                          data_prep_config: Optional[dict] = None) -> None:
    if train_data_path.startswith('dbfs:/'):
        validate_uc_path(train_data_path)
        if isinstance(data_prep_config, dict) and len(data_prep_config.keys()) > 0:
            validate_data_prep(train_data_path=train_data_path,
                               eval_data_path=eval_data_path,
                               data_prep_config=data_prep_config)
    elif "/" in train_data_path:  # assume HF dataset TODO state this assumption in docs
        validate_hf_dataset(train_data_path)
    else:
        validate_delta_table(train_data_path, "train_data_path")
    if model_registry_path:
        validate_model_registry_path(model_registry_path)
    if experiment_path:
        validate_experiment_path(experiment_path)
    if eval_data_path is None:
        pass
    elif eval_data_path.startswith('dbfs:/'):
        validate_uc_path(eval_data_path)
    elif "/" in eval_data_path:  # assume HF
        validate_hf_dataset(eval_data_path)
    else:
        validate_delta_table(eval_data_path, "eval_data_path")


def format_path(path: Union[str, None]) -> Union[str, None]:
    """
    Prepends `dbfs:` in front of paths that start with `/Volumes`.
    """
    if isinstance(path, str) and path.startswith('/Volumes'):
        return f'dbfs:{path}'
    else:
        return path



""" Common models
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict, Generic, Iterator, List, TypeVar

O = TypeVar('O', bound=type(dataclass))


class Object:

    @classmethod
    def from_mcli(cls, obj: Any) -> 'Object':
        raise NotImplementedError(f"from_mcli not implemented for {cls}")

    def _get_display_columns(self) -> Dict[str, str]:
        return {}


def generate_html_table(data: List[O], columns: Dict[str, str]):
    res = []
    res.append("<table border=\"1\" class=\"dataframe\">")

    # header
    res.append("<thead>")
    res.append("<tr style=\"text-align: right;\">")
    for col in columns.values():
        res.append(f"<th>{col}</th>")
    res.append("</tr>")
    res.append("</thead>")

    # body
    res.append("<tbody>")
    for row in data:
        res.append("<tr>")
        for col in columns:
            value = getattr(row, col, '')
            res.append(f"<td>{value}</td>")
        res.append("</tr>")
    res.append("</tbody>")

    res.append("</table>")
    return "\n".join(res)


class ObjectList(Generic[O]):
    """Common helper for list of objects
    """

    def __init__(self, data: List[O], obj_type: O):
        new_data = []
        for o in data:
            if isinstance(o, obj_type):
                new_data.append(o)
            else:
                new_data.append(obj_type.from_mcli(o))
        self.data = new_data
        self.type = obj_type

    def __iter__(self) -> Iterator[O]:
        return iter(self.data)

    def __getitem__(self, index: int) -> O:
        return self.data[index]

    def __len__(self) -> int:
        return len(self.data)

    @property
    def display_columns(self) -> Dict[str, str]:
        return self.type._get_display_columns(self.type)  # pylint: disable=protected-access

    def __str__(self):
        return '[' + ", ".join([str(i) for i in self.data]) + ']'

    def _repr_html_(self) -> str:
        return generate_html_table(self.data, self.display_columns)

    def to_pandas(self):
        try:
            # pylint: disable=import-outside-toplevel
            import pandas as pd  # type: ignore
        except ImportError as e:
            raise ImportError("Please install pandas to use this feature") from e

        cols = self.display_columns
        res = {col: [] for col in cols}
        for row in self.data:
            for col in cols:
                value = getattr(row, col)
                res[col].append(value)

        return pd.DataFrame(data=res)



"""Create a finetuning run"""

from typing import Dict, List, Optional

from mcli.api.finetuning_runs.api_create_finetuning_run import _db_create_finetuning_run

from databricks_genai.api.config import configure_request, get_me
from databricks_genai.api.finetuning.utils import format_path, validate_create_finetuning_run_inputs
from databricks_genai.types.finetuning import FinetuningRun


@configure_request
def create(
    model: str,
    train_data_path: str,
    model_registry_path: str,
    *,
    experiment_path: Optional[str] = "{}",
    task_type: Optional[str] = "INSTRUCTION_FINETUNE",
    eval_data_path: Optional[str] = None,
    eval_prompts: Optional[List[str]] = None,
    custom_weights_path: Optional[str] = None,
    training_duration: Optional[str] = None,
    learning_rate: Optional[float] = None,
    context_length: Optional[int] = None,
    validate_inputs: Optional[bool] = True,
    data_prep_config: Optional[Dict[str, str]] = None,
) -> FinetuningRun:
    """Create a finetuning run
    
    Args:
        model (str): The name of the Hugging Face model to use.
        train_data_path (str): The full remote location of your training data.
            The format of this data depends on the task type:
            - ``INSTRUCTION_FINETUNE``:  JSONL format, where each line is a 
                prompt and response JSON object, for example:
                {"prompt": "What is the capital of France?", "response": "Paris"}
            - ``CONTINUED_PRETRAIN``: A text file containing raw text data.
        model_registry_path (str): A Unity Catalog path where the model will
            be registered after training for easy deployment. 
        experiment_path (str, optional): The path to the MLflow experiment where
            the final model checkpoint will be saved. Defaults to the run name
        task_type (str, optional): The type of task to train for. Options:
            - ``INSTRUCTION_FINETUNE`` (default): Finetune a model with instructions
                relative to a specific task.
            - ``CONTINUED_PRETRAIN``: Continue pretraining a model using additional
                raw text data.
        eval_data_path (str, optional): The remote location of your evaluation data
            (if any). Defaults to no evaluation. Must follow the same format as
            ``train_data_path``.
        eval_prompts (List[str], optional): A list of prompt strings to generate 
            during evaluation. Results will be logged to the experiment every tim
            the model is checkpointed. Default is ``None`` (do not generate prompts).
        custom_weights_path (str, None) The remote location of a custom model checkpoint 
            to use for finetuning. If provided, these weights will be used instead of 
            the original pretrained weights of the model. This must be a Composer 
            checkpoint. Default is ``None``.
        training_duration: The total duration of your finetuning run. 
            This can be specified:
            - In batches (e.g. ``100ba``)
            - In epochs (e.g. ``10ep``)
            - In tokens (e.g. ``1_000_000tok``)
            Default is ``1ep``.
        learning_rate: The peak learning rate to use for finetuning. Default is ``5e-7``. 
        context_length: The maximum sequence length to use. This will be used to truncate
            any data that is too long. The default is the default for the provided Hugging 
            Face model. We do not support extending the context length beyond each model's
            default.
        validate_inputs: Whether to validate the access to input paths before submitting 
            the finetuning job. Default is ``True``.
        data_prep_config: Enable user to specify a clusterId for Spark data processing.
            This is required to support Delta table as an input for the finetuning API
            because we need to concatenate underlying Delta data files into a single location 
            and then convert to JSONL for IFT, and MDS for CPT.
    
    Returns:
        FinetuningRun: The finetuning run object that was created
    """
    databricks_username = get_me()
    experiment_tracker = {
        'mlflow': {
            'experiment_path': f'/Users/{databricks_username}/{experiment_path}',
            'model_registry_path': model_registry_path,
        }
    }

    train_data_path = format_path(train_data_path)
    eval_data_path = format_path(eval_data_path)
    custom_weights_path = format_path(custom_weights_path)
    save_folder = 'dbfs:/databricks/mlflow-tracking/{mlflow_experiment_id}/{mlflow_run_id}/artifacts'
    if validate_inputs:
        # don't validate experiment path if it's the default
        # TODO: create FinetuneConfig object for this SDK, so we can pass in object
        # instead of a list of params in the next line
        experiment_path_to_validate = experiment_tracker['mlflow'][
            'experiment_path'] if experiment_path != "{}" else None
        validate_create_finetuning_run_inputs(train_data_path, model_registry_path, experiment_path_to_validate,
                                              eval_data_path, data_prep_config)
    return FinetuningRun.from_mcli(
        _db_create_finetuning_run(
            model=model,
            train_data_path=train_data_path,
            task_type=task_type,
            eval_data_path=eval_data_path,
            eval_prompts=eval_prompts,
            data_prep_config=data_prep_config,
            custom_weights_path=custom_weights_path,
            training_duration=training_duration,
            learning_rate=learning_rate,
            context_length=context_length,
            save_folder=save_folder,
            experiment_tracker=experiment_tracker,
            disable_credentials_check=True,
        ))



"""
A Databricks finetuning run
"""

from dataclasses import dataclass
from datetime import datetime
from typing import Any, Dict, List, Optional

from mcli import Finetune as McliFinetune
from mcli.api.model.run_event import FormattedRunEvent as McliFinetuneEvent
from mcli.utils.utils_run_status import RunStatus

from databricks_genai.types.common import Object


@dataclass
class FinetuningEvent(Object):
    """
    An event that occurs during a finetuning run

    Args:
        type: The type of event
        time: The time the event occurred
        message: The message associated with the event
    """

    type: str
    time: datetime
    message: str

    @classmethod
    def from_mcli(cls, obj: McliFinetuneEvent) -> 'FinetuningEvent':
        return cls(
            type=obj.event_type,
            time=obj.event_time,
            message=obj.event_message,
        )

    def _get_display_columns(self) -> Dict[str, str]:
        return {
            'type': 'Type',
            'time': 'Time',
            'message': 'Message',
        }


@dataclass
class FinetuningRun:
    """A Databricks finetuning run
    
    Args:
        name: The name of the finetuning run
        created_by: The user email of who created the run
        model: The model to finetune
        train_data_path: The path to the training data
        model_registry_path: The path to the model registry
        experiment_path: The path to save the MLflow experiment
        created_at: The time the run was created
        started_at: The time the run was started
        estimated_end_time: The estimated time the run will complete
        completed_at: The time the run was completed
    """

    name: str
    status: RunStatus
    created_by: str

    # User inputs
    model: str
    train_data_path: str
    model_registry_path: str
    experiment_path: Optional[str] = None
    data_prep_config: Optional[Dict[str, str]] = None

    # Lifecycle
    created_at: Optional[datetime] = None
    started_at: Optional[datetime] = None
    estimated_end_time: Optional[datetime] = None
    completed_at: Optional[datetime] = None

    # hyperparameters: Hyperparameters ## todo

    @property
    def submitted_config(self) -> Dict[str, Any]:
        """The config originally submitted to the API to create this run"""
        return {
            'model': self.model,
            'train_data_path': self.train_data_path,
            'model_registry_path': self.model_registry_path,
            'experiment_path': self.experiment_path,
            # TODO: add the other inputs from create here
        }

    def _get_display_columns(self) -> Dict[str, str]:
        return {
            'name': 'Name',
            'status': 'Status',
            'created_by': 'Created By',
            'model': 'Model',
        }

    @classmethod
    def from_mcli(
        cls,
        obj: McliFinetune,
    ) -> 'FinetuningRun':

        # TODO: These should be top level properties, for now pull from the original yaml
        submitted_config = obj.submitted_config
        mlflow_tracker = submitted_config.experiment_tracker["mlflow"]
        # The experiment path is auto-formatted to be '/Users/username/experiment_path'
        _, _, _, *experiment_path = mlflow_tracker['experiment_path'].split('/')
        experiment_path = '/'.join(experiment_path)
        model_registry_path = mlflow_tracker.get('model_registry_path')

        return cls(
            name=obj.name,
            status=obj.status,
            created_by=obj.created_by,
            model=obj.model,
            train_data_path=obj.train_data_path,
            model_registry_path=model_registry_path,
            experiment_path=experiment_path,
            data_prep_config=submitted_config.data_prep_config,
            created_at=obj.created_at,
            started_at=obj.started_at,
            completed_at=obj.completed_at,
            estimated_end_time=obj.estimated_end_time,
        )

    def refresh(self) -> 'FinetuningRun':
        """Refetches the finetuning run from the API

        Returns:
            FinetuningRun: The updated finetuning run
        """

        # pylint: disable=import-outside-toplevel, cyclic-import
        from databricks_genai.api.finetuning.get import get
        return get(self)

    def cancel(self) -> int:
        """Cancel the finetuning run
        
        Returns:
            int: Will return 1 if the run was cancelled, 0 if it was already cancelled
        """

        # pylint: disable=import-outside-toplevel, cyclic-import
        from databricks_genai.api.finetuning.cancel import cancel
        return cancel(self)

    def delete(self) -> int:
        """Delete the finetuning run
        
        Returns:
            int: Will return 1 if the run was deleted, 0 if it was already deleted
        """

        # pylint: disable=import-outside-toplevel, cyclic-import
        from databricks_genai.api.finetuning.delete import delete
        return delete(self)

    def get_events(self) -> List[FinetuningEvent]:
        """Get events for the finetuning run
        
        Returns:
            List[FinetuningEvent]: A list of finetuning events. Each event has an event 
                type, time, and message.
        """

        # pylint: disable=import-outside-toplevel, cyclic-import
        from databricks_genai.api.finetuning.get_events import get_events
        return get_events(self)



"""Cancel a finetuning run"""

from typing import List, Union

from mcli import stop_finetuning_runs

from databricks_genai.api.config import configure_request
from databricks_genai.errors import DatabricksGenAIRequestError
from databricks_genai.types.finetuning import FinetuningRun


@configure_request
def cancel(finetuning_runs: Union[str, FinetuningRun, List[str], List[FinetuningRun]]) -> int:
    """Cancel a finetuning run or list of finetuning runs without deleting them
    
    Args:
        finetuning_runs (Union[str, FinetuningRun, List[str], List[FinetuningRun]]): The
            finetuning run(s) to cancel. Can be a single run or a list of runs.
    
    Returns:
        int: The number of finetuning runs cancelled
    """

    if not finetuning_runs:
        raise DatabricksGenAIRequestError('Must provide finetuning run(s) to cancel')

    if isinstance(finetuning_runs, (str, FinetuningRun)):
        finetuning_runs = [finetuning_runs]

    res = stop_finetuning_runs(finetuning_runs=[f if isinstance(f, str) else f.name for f in finetuning_runs])
    return len(res)
